/*
 * Wireless Extensions for RNDIS Networking Links
 * Copyright (C) 2007 by Bjorge Dijkstra (bjorge@gmx.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/wireless.h>
#include <linux/if_arp.h>
#include <net/iw_handler.h>


/* codes for "status" field of completion messages */
#define RNDIS_STATUS_SUCCESS		ccpu2(0x00000000)
#define RNDIS_STATUS_FAILURE		ccpu2(0xc0000001)
#define RNDIS_STATUS_NOT_SUPPORTED	ccpu2(0xc00000bb)
#define RNDIS_STATUS_ADAPTER_NOT_READY	ccpu2(0xc0010011)
#define RNDIS_STATUS_INVALID_DATA	ccpu2(0xc0010015)

#define RNDIS_STATUS_MEDIA_CONNECT	ccpu2(0x4001000b)
#define RNDIS_STATUS_MEDIA_DISCONNECT	ccpu2(0x4001000c)

/* various RNDIS OID defs */
#define OID_GEN_SUPPORTED_LIST			ccpu2(0x00010101)
#define OID_GEN_LINK_SPEED			ccpu2(0x00010107)

#define OID_802_3_PERMANENT_ADDRESS		ccpu2(0x01010101)
#define OID_802_3_CURRENT_ADDRESS		ccpu2(0x01010102)
#define OID_802_3_MULTICAST_LIST        	ccpu2(0x01010103)

#define OID_802_11_BSSID			ccpu2(0x0d010101)
#define OID_802_11_SSID				ccpu2(0x0d010102)
#define OID_802_11_INFRASTRUCTURE_MODE		ccpu2(0x0d010108)
#define OID_802_11_ADD_WEP			ccpu2(0x0d010113)
#define OID_802_11_REMOVE_WEP			ccpu2(0x0d010114)
#define OID_802_11_DISASSOCIATE			ccpu2(0x0d010115)
#define OID_802_11_AUTHENTICATION_MODE		ccpu2(0x0d010118)
#define OID_802_11_PRIVACY_FILTER		ccpu2(0x0d010119)
#define OID_802_11_BSSID_LIST_SCAN		ccpu2(0x0d01011a)
#define OID_802_11_ENCRYPTION_STATUS		ccpu2(0x0d01011b)
#define OID_802_11_ADD_KEY			ccpu2(0x0d01011d)
#define OID_802_11_REMOVE_KEY			ccpu2(0x0d01011e)
#define OID_802_11_PMKID			ccpu2(0x0d010123)
#define OID_802_11_NETWORK_TYPES_SUPPORTED	ccpu2(0x0d010203)
#define OID_802_11_NETWORK_TYPE_IN_USE		ccpu2(0x0d010204)
#define OID_802_11_TX_POWER_LEVEL		ccpu2(0x0d010205)
#define OID_802_11_RSSI				ccpu2(0x0d010206)
#define OID_802_11_RSSI_TRIGGER			ccpu2(0x0d010207)
#define OID_802_11_FRAGMENTATION_THRESHOLD	ccpu2(0x0d010209)
#define OID_802_11_RTS_THRESHOLD		ccpu2(0x0d01020a)
#define OID_802_11_SUPPORTED_RATES		ccpu2(0x0d01020e)
#define OID_802_11_CONFIGURATION		ccpu2(0x0d010211)
#define OID_802_11_BSSID_LIST			ccpu2(0x0d010217)



/* NDIS data structures. Taken from wpa_supplicant driver_ndis.c
 * slightly modified for datatype endianess, etc
 */
#define NDIS_802_11_LENGTH_SSID 32
#define NDIS_802_11_LENGTH_RATES 8
#define NDIS_802_11_LENGTH_RATES_EX 16

typedef u8 NDIS_802_11_MAC_ADDRESS[6];

typedef struct NDIS_802_11_SSID {
	__le32 SsidLength;
	u8 Ssid[NDIS_802_11_LENGTH_SSID];
} NDIS_802_11_SSID;

typedef __le32 NDIS_802_11_RSSI;  /* signed type */

typedef enum NDIS_802_11_NETWORK_TYPE {
	Ndis802_11FH,
	Ndis802_11DS,
	Ndis802_11OFDM5,
	Ndis802_11OFDM24,
	Ndis802_11NetworkTypeMax
} NDIS_802_11_NETWORK_TYPE;

typedef struct NDIS_802_11_CONFIGURATION_FH {
	__le32 Length;
	__le32 HopPattern;
	__le32 HopSet;
	__le32 DwellTime;
} NDIS_802_11_CONFIGURATION_FH;

typedef struct NDIS_802_11_CONFIGURATION {
	__le32 Length;
	__le32 BeaconPeriod;
	__le32 ATIMWindow;
	__le32 DSConfig;
	NDIS_802_11_CONFIGURATION_FH FHConfig;
} NDIS_802_11_CONFIGURATION;

typedef enum NDIS_802_11_NETWORK_INFRASTRUCTURE {
	Ndis802_11IBSS,
	Ndis802_11Infrastructure,
	Ndis802_11AutoUnknown,
	Ndis802_11InfrastructureMax
} NDIS_802_11_NETWORK_INFRASTRUCTURE;

typedef enum NDIS_802_11_AUTHENTICATION_MODE {
	Ndis802_11AuthModeOpen,
	Ndis802_11AuthModeShared,
	Ndis802_11AuthModeAutoSwitch,
	Ndis802_11AuthModeWPA,
	Ndis802_11AuthModeWPAPSK,
	Ndis802_11AuthModeWPANone,
	Ndis802_11AuthModeWPA2,
	Ndis802_11AuthModeWPA2PSK,
	Ndis802_11AuthModeMax
} NDIS_802_11_AUTHENTICATION_MODE;

typedef enum NDIS_802_11_WEP_STATUS {
	Ndis802_11WEPEnabled,
	Ndis802_11Encryption1Enabled = Ndis802_11WEPEnabled,
	Ndis802_11WEPDisabled,
	Ndis802_11EncryptionDisabled = Ndis802_11WEPDisabled,
	Ndis802_11WEPKeyAbsent,
	Ndis802_11Encryption1KeyAbsent = Ndis802_11WEPKeyAbsent,
	Ndis802_11WEPNotSupported,
	Ndis802_11EncryptionNotSupported = Ndis802_11WEPNotSupported,
	Ndis802_11Encryption2Enabled,
	Ndis802_11Encryption2KeyAbsent,
	Ndis802_11Encryption3Enabled,
	Ndis802_11Encryption3KeyAbsent
} NDIS_802_11_WEP_STATUS, NDIS_802_11_ENCRYPTION_STATUS;

typedef enum NDIS_802_11_PRIVACY_FILTER {
	Ndis802_11PrivFilterAcceptAll,
	Ndis802_11PrivFilter8021xWEP
} NDIS_802_11_PRIVACY_FILTER;

typedef u8 NDIS_802_11_RATES[NDIS_802_11_LENGTH_RATES];
typedef u8 NDIS_802_11_RATES_EX[NDIS_802_11_LENGTH_RATES_EX];

typedef struct NDIS_WLAN_BSSID_EX {
	__le32 Length;
	NDIS_802_11_MAC_ADDRESS MacAddress; /* BSSID */
	u8 Reserved[2];
	NDIS_802_11_SSID Ssid;
	__le32 Privacy;
	NDIS_802_11_RSSI Rssi;
	NDIS_802_11_NETWORK_TYPE NetworkTypeInUse;
	NDIS_802_11_CONFIGURATION Configuration;
	NDIS_802_11_NETWORK_INFRASTRUCTURE InfrastructureMode;
	NDIS_802_11_RATES_EX SupportedRates;
	__le32 IELength;
	u8 IEs[0];
} NDIS_WLAN_BSSID_EX;

typedef struct NDIS_802_11_BSSID_LIST_EX {
	__le32 NumberOfItems;
	NDIS_WLAN_BSSID_EX Bssid[0];
} NDIS_802_11_BSSID_LIST_EX;

typedef struct NDIS_802_11_FIXED_IEs {
	u8 Timestamp[8];
	__le16 BeaconInterval;
	__le16 Capabilities;
} NDIS_802_11_FIXED_IEs;

typedef struct NDIS_802_11_WEP {
	__le32 Length;
	__le32 KeyIndex;
	__le32 KeyLength;
	u8 KeyMaterial[0];
} NDIS_802_11_WEP;

typedef __le32 NDIS_802_11_KEY_INDEX;
typedef __le64 NDIS_802_11_KEY_RSC;

typedef struct NDIS_802_11_KEY {
	__le32 Length;
	__le32 KeyIndex;
	__le32 KeyLength;
	NDIS_802_11_MAC_ADDRESS BSSID;
	NDIS_802_11_KEY_RSC KeyRSC;
	u8 KeyMaterial[32];
} NDIS_802_11_KEY;

typedef struct NDIS_802_11_REMOVE_KEY {
	__le32 Length;
	__le32 KeyIndex;
	NDIS_802_11_MAC_ADDRESS BSSID;
} NDIS_802_11_REMOVE_KEY;

typedef struct NDIS_802_11_AI_REQFI {
	__le16 Capabilities;
	__le16 ListenInterval;
	NDIS_802_11_MAC_ADDRESS CurrentAPAddress;
} NDIS_802_11_AI_REQFI;

typedef struct NDIS_802_11_AI_RESFI {
	__le16 Capabilities;
	__le16 StatusCode;
	__le16 AssociationId;
} NDIS_802_11_AI_RESFI;

typedef struct NDIS_802_11_ASSOCIATION_INFORMATION {
	__le32 Length;
	__le16 AvailableRequestFixedIEs;
	NDIS_802_11_AI_REQFI RequestFixedIEs;
	__le32 RequestIELength;
	__le32 OffsetRequestIEs;
	__le16 AvailableResponseFixedIEs;
	NDIS_802_11_AI_RESFI ResponseFixedIEs;
	__le32 ResponseIELength;
	__le32 OffsetResponseIEs;
} NDIS_802_11_ASSOCIATION_INFORMATION;

typedef struct NDIS_802_11_AUTHENTICATION_ENCRYPTION {
	NDIS_802_11_AUTHENTICATION_MODE AuthModeSupported;
	NDIS_802_11_ENCRYPTION_STATUS EncryptStatusSupported;
} NDIS_802_11_AUTHENTICATION_ENCRYPTION;

typedef struct NDIS_802_11_CAPABILITY {
	__le32 Length;
	__le32 Version;
	__le32 NoOfPMKIDs;
	__le32 NoOfAuthEncryptPairSupported;
	NDIS_802_11_AUTHENTICATION_ENCRYPTION
		AuthenticationEncryptionSupported[0];
} NDIS_802_11_CAPABILITY;

typedef u8 NDIS_802_11_PMKID_VALUE[16];

typedef struct BSSID_INFO {
	NDIS_802_11_MAC_ADDRESS BSSID;
	NDIS_802_11_PMKID_VALUE PMKID;
} BSSID_INFO;

typedef struct NDIS_802_11_PMKID {
	__le32 Length;
	__le32 BSSIDInfoCount;
	BSSID_INFO BSSIDInfo[0];
} NDIS_802_11_PMKID;

typedef enum NDIS_802_11_STATUS_TYPE {
	Ndis802_11StatusType_Authentication,
	Ndis802_11StatusType_PMKID_CandidateList = 2,
	Ndis802_11StatusTypeMax
} NDIS_802_11_STATUS_TYPE;

typedef struct NDIS_802_11_STATUS_INDICATION {
	NDIS_802_11_STATUS_TYPE StatusType;
} NDIS_802_11_STATUS_INDICATION;

typedef struct PMKID_CANDIDATE {
	NDIS_802_11_MAC_ADDRESS BSSID;
	__le32 Flags;
} PMKID_CANDIDATE;

#define NDIS_802_11_PMKID_CANDIDATE_PREAUTH_ENABLED 0x01

typedef struct NDIS_802_11_PMKID_CANDIDATE_LIST {
	__le32 Version;
	__le32 NumCandidates;
	PMKID_CANDIDATE CandidateList[0];
} NDIS_802_11_PMKID_CANDIDATE_LIST;

typedef struct NDIS_802_11_AUTHENTICATION_REQUEST {
	__le32 Length;
	NDIS_802_11_MAC_ADDRESS Bssid;
	__le32 Flags;
} NDIS_802_11_AUTHENTICATION_REQUEST;

#define NDIS_802_11_AUTH_REQUEST_REAUTH			0x01
#define NDIS_802_11_AUTH_REQUEST_KEYUPDATE		0x02
#define NDIS_802_11_AUTH_REQUEST_PAIRWISE_ERROR		0x06
#define NDIS_802_11_AUTH_REQUEST_GROUP_ERROR		0x0E


/*
 *  private data
 */
#define NET_TYPE_11FB	0

#define CAP_MODE_80211A		1
#define CAP_MODE_80211B		2
#define CAP_MODE_80211G		4
#define CAP_MODE_MASK		7

/* RNDIS device private data */
struct rndis_private {
	char name[32];

	struct iw_statistics iwstats;
	int  nick_len;
	char nick[32];
	u8   bssid[6];

	int caps;
	int mode;

	/* general encryption stuff */
	int privacy;

	/* wep stuff */
	int  wep_def_key;
	int  wep_enabled;
	char wep_keys[4][13];
	int  wep_key_len[4];

	/* wpa stuff */
	int wpa_enabled;
	int wpa_version;
	int wpa_keymgmt;
	int wpa_authalg;
	int wpa_ie_len;
	u8 *wpa_ie;
	int wpa_cipher_pair;
	int wpa_cipher_group;
};


/* translate error code */
static int rndis_error_status(__le32 rndis_status)
{
	int ret = -EINVAL;
	switch (rndis_status) {
	case RNDIS_STATUS_SUCCESS:
		ret = 0;
		break;
	case RNDIS_STATUS_FAILURE:
	case RNDIS_STATUS_INVALID_DATA:
		ret = -EINVAL;
		break;
	case RNDIS_STATUS_NOT_SUPPORTED:
		ret = -ENOTSUPP;
		break;
	}
	return ret;
}


static int rndis_query_oid(struct usbnet *dev, __le32 oid, void *data, int *len)
{
	union {
		void			*buf;
		struct rndis_msg_hdr	*header;
		struct rndis_query	*get;
		struct rndis_query_c	*get_c;
	} u;
	int ret, buflen;

	buflen = *len + sizeof(*u.get);
	if (buflen < CONTROL_BUFFER_SIZE)
		buflen = CONTROL_BUFFER_SIZE;
	u.buf = kmalloc(buflen, GFP_KERNEL);
	if (!u.buf)
		return -ENOMEM;
	memset(u.get, 0, sizeof *u.get);
	u.get->msg_type = RNDIS_MSG_QUERY;
	u.get->msg_len = ccpu2(sizeof *u.get);
	u.get->oid = oid;
	ret = rndis_command(dev, u.header);

	if (ret == 0) {
		ret = le32_to_cpu(u.get_c->len);
		*len = (*len > ret) ? ret : *len;
		memcpy(data, u.buf + le32_to_cpu(u.get_c->offset) + 8, *len);
		ret = rndis_error_status(u.get_c->status);
	}
	kfree(u.buf);
	return ret;
}


static int rndis_set_oid(struct usbnet *dev, __le32 oid, void *data, int len)
{
	union {
		void			*buf;
		struct rndis_msg_hdr	*header;
		struct rndis_set	*set;
		struct rndis_set_c	*set_c;
	} u;
	int ret, buflen;

	buflen = len + sizeof(*u.set);
	if (buflen < CONTROL_BUFFER_SIZE)
		buflen = CONTROL_BUFFER_SIZE;
	u.buf = kmalloc(buflen, GFP_KERNEL);
	if (!u.buf)
		return -ENOMEM;

	memset(u.set, 0, sizeof *u.set);
	u.set->msg_type = RNDIS_MSG_SET;
	u.set->msg_len = cpu_to_le32(sizeof(*u.set) + len);
	u.set->oid = oid;
	u.set->len = cpu_to_le32(len);
	u.set->offset = ccpu2(sizeof(*u.set) - 8);
	u.set->handle = ccpu2(0);
	memcpy(u.buf + sizeof(*u.set), data, len);

	ret = rndis_command(dev, u.header);

	if (ret == 0) {
	    ret = rndis_error_status(u.set_c->status);
	}
	
	kfree(u.buf);
	return ret;
}


/*
 * wireless extension handlers
 */

static int rndis_iw_commit(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data *wrqu, char * extra)
{
	// dummy op
	return 0;
}



static int rndis_iw_get_range(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data *wrqu, char * extra)
{
	struct iw_range * range = (struct iw_range*)extra;
	struct usbnet * usbdev = dev->priv;
	int len, ret, i;
	u8 rates[8];

	devdbg(usbdev, "SIOCGIWRANGE");

	// clear iw_range struct
	memset(range, 0, sizeof(*range));
	wrqu->data.length = sizeof(*range);

	// estimated max througput in bps
	range->throughput = 54000000;

	len = sizeof(rates);
	ret = rndis_query_oid(usbdev, OID_802_11_SUPPORTED_RATES, &rates,
									&len);
	if (ret == 0) {
		for (i = 0; i < len; i++) {
			range->bitrate[i] = (rates[i] & 0x7f) * 500000;
			if (rates[i] == 0)
				break;
		}
		range->num_bitrates = i;
	} else
		range->num_bitrates = 0;

	range->min_rts = 0;
	range->max_rts = 2347;
	range->min_frag = 256;
	range->max_frag = 2346;

	// quality ranges are just guesses...
	// msdn documentation for OID_802_11_RSSI indicates normal range
	// of -200dBm to -10dBm for signal strength
	range->max_qual.qual = -10;
	range->max_qual.level = -10;
	range->max_qual.noise = 0;
	range->max_qual.updated = IW_QUAL_QUAL_UPDATED
				| IW_QUAL_LEVEL_UPDATED
				| IW_QUAL_DBM
				| IW_QUAL_NOISE_INVALID;

	range->avg_qual.qual = -80;
	range->avg_qual.level = -80;
	range->avg_qual.noise = 0;
	range->avg_qual.updated = IW_QUAL_QUAL_UPDATED
				| IW_QUAL_LEVEL_UPDATED
				| IW_QUAL_DBM
				| IW_QUAL_NOISE_INVALID;

	range->we_version_compiled = WIRELESS_EXT;
	range->we_version_source = WIRELESS_EXT;

	range->enc_capa = IW_ENC_CAPA_WPA | IW_ENC_CAPA_WPA2 | 
			IW_ENC_CAPA_CIPHER_TKIP | IW_ENC_CAPA_CIPHER_CCMP;
	return 0;
}



static int rndis_iw_get_name(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data *wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	struct rndis_private * priv = usbdev->rndis_wext_priv;

	strcpy(wrqu->name, priv->name);
	return 0;
}

static int rndis_iw_set_essid(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data *wrqu, char *essid)
{
	NDIS_802_11_SSID ssid;
	int length = wrqu->essid.length;
	struct usbnet * usbdev = dev->priv;

	devdbg(usbdev, "SIOCSIWESSID: [flags:%d,len:%d] '%32s'",
		wrqu->essid.flags, wrqu->essid.length, essid);

	if (length > NDIS_802_11_LENGTH_SSID)
		length = NDIS_802_11_LENGTH_SSID;

	ssid.SsidLength = cpu_to_le32(length);
	if (length > 0) {
		memcpy(ssid.Ssid, essid, length);
	} else {
		memset(ssid.Ssid, 0, NDIS_802_11_LENGTH_SSID);
	}
	return rndis_set_oid(usbdev, OID_802_11_SSID, &ssid, sizeof(ssid));
}



static int rndis_iw_get_essid(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data *wrqu, char * essid)
{
	NDIS_802_11_SSID ssid;
	struct usbnet * usbdev = dev->priv;
	int ret, length;

	length = sizeof(ssid);
	ret = rndis_query_oid(usbdev, OID_802_11_SSID, &ssid, &length);

	if (ret == 0 && ssid.SsidLength) {
		wrqu->essid.flags = 1;
		wrqu->essid.length = le32_to_cpu(ssid.SsidLength);
		memcpy(essid, ssid.Ssid, wrqu->essid.length);
	} else {
		memset(essid, 0, sizeof(NDIS_802_11_LENGTH_SSID));
		wrqu->essid.flags = 0;
		wrqu->essid.length = 0;
	}
	devdbg(usbdev, "SIOCGIWESSID: %s", essid);
	return ret;
}


static int rndis_iw_get_bssid(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	unsigned char bssid[ETH_ALEN];
	int ret, len;

	len = ETH_ALEN;
	ret = rndis_query_oid(usbdev, OID_802_11_BSSID, &bssid, &len);

	wrqu->ap_addr.sa_family = ARPHRD_ETHER;
	if (ret == 0) {
		devdbg(usbdev, "SIOCGIWAP: %02x:%02x:%02x:%02x:%02x:%02x",
			bssid[0], bssid[1], bssid[2],
			bssid[3], bssid[4], bssid[5]);
		memcpy(wrqu->ap_addr.sa_data, bssid, ETH_ALEN);
	} else {
		devdbg(usbdev, "SIOCGIWAP: <not associated>");
		memset(wrqu->ap_addr.sa_data, 0, ETH_ALEN);
	}

	return ret;
}


static int rndis_iw_set_bssid(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	u8 * bssid = (u8*)wrqu->ap_addr.sa_data;

	devdbg(usbdev, "SIOCSIWAP: %02x:%02x:%02x:%02x:%02x:%02x",
				bssid[0], bssid[1], bssid[2],
				bssid[3], bssid[4], bssid[5]);
			
	return rndis_set_oid(usbdev, OID_802_11_BSSID, bssid, ETH_ALEN);
}



static int rndis_iw_set_auth(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct iw_param * p = &wrqu->param;
	struct usbnet * usbdev = dev->priv;
	struct rndis_private * priv = usbdev->rndis_wext_priv;
	int ret = -ENOTSUPP;
	int auth = Ndis802_11AuthModeOpen;
	int enc = Ndis802_11EncryptionDisabled;
	__le32 tmp;

	switch (p->flags & IW_AUTH_INDEX) {
	case IW_AUTH_WPA_VERSION:
		devdbg(usbdev, "SIOCSIWAUTH: WPA_VERSION, %08x", p->value);
		priv->wpa_version |= p->value;
		ret = 0;
		break;
	
	case IW_AUTH_CIPHER_PAIRWISE:
		devdbg(usbdev, "SIOCSIWAUTH: CIPHER_PAIRWISE, %08x",p->value);
		priv->wpa_cipher_pair = p->value;
		ret = 0;
		break;
	
	case IW_AUTH_CIPHER_GROUP:
		devdbg(usbdev, "SIOCSIWAUTH: CIPHER_GROUP, %08x", p->value);
		priv->wpa_cipher_group = p->value;
		ret = 0;
		break;

	case IW_AUTH_KEY_MGMT:
		devdbg(usbdev, "SIOCSIWAUTH: KEY_MGMT, %08x", p->value);
		priv->wpa_keymgmt = p->value;
		ret = 0;
		break;
	
	case IW_AUTH_TKIP_COUNTERMEASURES:
		devdbg(usbdev, "SIOCSIWAUTH: TKIP_COUNTERMEASURES, %08x",
								p->value);
		ret = 0;
		break;
	
	case IW_AUTH_DROP_UNENCRYPTED:
		devdbg(usbdev, "SIOCSIWAUTH: DROP_UNENCRYPTED, %08x", p->value);
		ret = 0;
		break;

	case IW_AUTH_80211_AUTH_ALG:
		devdbg(usbdev, "SIOCSIWAUTH: 80211_AUTH_ALG, %08x", p->value);
		priv->wpa_authalg = p->value;
		ret = 0;
		break;
	
	case IW_AUTH_WPA_ENABLED:
		devdbg(usbdev, "SIOCSIWAUTH: WPA_ENABLED, %08x", p->value);
		if (!p->value) {
			priv->wpa_version |= IW_AUTH_WPA_VERSION_DISABLED;
		} else {
			priv->wpa_version &= ~IW_AUTH_WPA_VERSION_DISABLED;
		}
		ret = 0;
		break;

	case IW_AUTH_RX_UNENCRYPTED_EAPOL:
		devdbg(usbdev, "SIOCSIWAUTH: RX_UNENCRYPTED_EAPOL, %08x",
								p->value);
		ret = 0;
		break;
	
	case IW_AUTH_ROAMING_CONTROL:
		devdbg(usbdev, "SIOCSIWAUTH: ROAMING_CONTROL, %08x", p->value);
		ret = 0;
		break;
	
	case IW_AUTH_PRIVACY_INVOKED:
		devdbg(usbdev, "SIOCSIWAUTH: PRIVACY_INVOKED, %08x", p->value);

		priv->privacy = p->value;

		if (priv->privacy) {
			switch (priv->wpa_keymgmt) {
			case IW_AUTH_KEY_MGMT_PSK:
				if (priv->wpa_version == 0) {
					auth = Ndis802_11AuthModeShared;
					enc  = Ndis802_11Encryption1Enabled;
				} else if (priv->wpa_version ==
						IW_AUTH_WPA_VERSION_WPA) {
					auth = Ndis802_11AuthModeWPAPSK;
					enc  = Ndis802_11Encryption2Enabled;
				} else if (priv->wpa_version ==
						IW_AUTH_WPA_VERSION_WPA2) {
					auth = Ndis802_11AuthModeWPA2PSK;
					enc  = Ndis802_11Encryption3Enabled;
				} else {
					devdbg(usbdev, "weird wpa_version %08x",
							priv->wpa_version);
				}
				break;

			case IW_AUTH_KEY_MGMT_802_1X:
				if (priv->wpa_version == 0) {
					auth = Ndis802_11AuthModeShared;
					enc  = Ndis802_11Encryption1Enabled;
				} else if (priv->wpa_version ==
						IW_AUTH_WPA_VERSION_WPA) {
					auth = Ndis802_11AuthModeWPA;
					enc  = Ndis802_11Encryption2Enabled;
				} else if (priv->wpa_version ==
						IW_AUTH_WPA_VERSION_WPA2) {
					auth = Ndis802_11AuthModeWPA2;
					enc  = Ndis802_11Encryption3Enabled;
				} else {
					devdbg(usbdev, "weird wpa_version %08x",
							priv->wpa_version);
				}
				break;
			}
		} else {
			auth = Ndis802_11AuthModeOpen;
			enc  = Ndis802_11EncryptionDisabled;
		}
		tmp = cpu_to_le32(auth);
		rndis_set_oid(usbdev, OID_802_11_AUTHENTICATION_MODE, &tmp, 
								sizeof(tmp));
		tmp = cpu_to_le32(enc);
		rndis_set_oid(usbdev, OID_802_11_ENCRYPTION_STATUS, &tmp,
								sizeof(tmp));
		ret = 0;
		break;

	default:
		devdbg(usbdev, "SIOCSIWAUTH: UKNOWN  %08x, %08x",
			p->flags & IW_AUTH_INDEX, p->value);
	}
	return ret;
}

static int rndis_iw_get_mode(struct net_device *dev, struct iw_request_info *info,
                           union iwreq_data *wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	struct rndis_private * priv = usbdev->rndis_wext_priv;

	switch (priv->mode) {
	case Ndis802_11IBSS:
		wrqu->mode = IW_MODE_ADHOC;
		break;
	case Ndis802_11Infrastructure:
		wrqu->mode = IW_MODE_INFRA;
		break;
	case Ndis802_11AutoUnknown:
		wrqu->mode = IW_MODE_AUTO;
		break;
	}
	devdbg(usbdev, "SIOCGIWMODE: %08x", wrqu->mode);
        return 0;
}


static int rndis_iw_set_mode(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	struct rndis_private * priv = usbdev->rndis_wext_priv;
	int ret = -EINVAL;
	__le32 tmp;

	devdbg(usbdev, "SIOCSIWMODE: %08x, %08x", param->flags, param->value);

	switch (wrqu->mode) {
	case IW_MODE_ADHOC:
		priv->mode = Ndis802_11IBSS;
		break;
	
	case IW_MODE_INFRA:
		priv->mode = Ndis802_11Infrastructure;
		break; 

	case IW_MODE_AUTO:
		priv->mode = Ndis802_11AutoUnknown;
		break;
	}
	tmp = cpu_to_le32(priv->mode);
	ret = rndis_set_oid(usbdev, OID_802_11_INFRASTRUCTURE_MODE, &tmp,
								sizeof(tmp));
	return ret;
}


static int rndis_iw_set_encode(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct iw_point * param = &wrqu->encoding;
	struct usbnet * usbdev = dev->priv;
	struct rndis_private * priv = usbdev->rndis_wext_priv;
	int ret = 0;
	NDIS_802_11_WEP	* wepkey;
	u32	keylength;
	int 	index;
	int 	auth =  Ndis802_11AuthModeOpen;
	int	enc = Ndis802_11EncryptionDisabled;
	__le32	tmp;

	devdbg(usbdev, "SIOCSIWENCODE: %08x, %08x",
		param->flags, param->length);

	// get memory for max size wep key
	wepkey = kmalloc(sizeof(*wepkey) + 13, GFP_KERNEL);
	if (!wepkey)
		return -ENOMEM;

	index = (param->flags & IW_ENCODE_INDEX) - 1;

	if (param->length > 0) {
		int current_index = priv->wep_def_key;
		/* Check the size of the key */
		if (param->length > 13) {
			ret = -EINVAL;
			goto out;
		}
		/* Check the index (none -> use current) */
		if (index < 0 || index >= 4)
			index = current_index;
		else
			priv->wep_def_key = index;
		/* Set the length */
		if (param->length > 5)
			priv->wep_key_len[index] = 13;
		else
			if (param->length > 0)
				priv->wep_key_len[index] = 5;
			else
				/* Disable the key */
				priv->wep_key_len[index] = 0;
		/* Check if the key is not marked as invalid */
		if(!(param->flags & IW_ENCODE_NOKEY)) {
			/* Cleanup */
			memset(priv->wep_keys[index], 0, 13);
			/* Copy the key in the driver */
			memcpy(priv->wep_keys[index], extra, param->length);
		}
		/* WE specify that if a valid key is set, encryption
		 * should be enabled (user may turn it off later)
		 * This is also how "iwconfig ethX key on" works */
		if (index == current_index &&
		    priv->wep_key_len[index] > 0) {
			priv->wep_enabled = 1;
		}
	} else {
		if (index >= 0 && index < 4) {
			priv->wep_def_key = index;
		} else
			/* Don't complain if only change the mode */
			if (!param->flags & IW_ENCODE_MODE) {
				ret = -EINVAL;
				goto out;
			}
	}

	if (index >= 0 && index < 4) {
		keylength = priv->wep_key_len[index];
		memcpy(wepkey->KeyMaterial, priv->wep_keys[index], keylength);
		wepkey->Length = cpu_to_le32(sizeof(*wepkey) + keylength);
		wepkey->KeyIndex = cpu_to_le32(index);
		if (index == 0)
			wepkey->KeyIndex = cpu_to_le32(1 << 31);
		wepkey->KeyLength = cpu_to_le32(keylength);
		ret = rndis_set_oid(usbdev, OID_802_11_ADD_WEP, wepkey,
						sizeof(*wepkey) + keylength);
	}


	/* Read the flags */
	if (param->flags & IW_ENCODE_DISABLED) {
		priv->wep_enabled = 0;
	} else {
		priv->wep_enabled = 1;
	}

	if (param->flags & IW_ENCODE_RESTRICTED)
		auth = Ndis802_11AuthModeShared;
	if (param->flags & IW_ENCODE_OPEN)
		auth = Ndis802_11AuthModeOpen;

	if (priv->wep_enabled) {
		if (param->flags & IW_ENCODE_MODE) {
			enc = Ndis802_11WEPEnabled;
		}
	} else {
		enc = Ndis802_11WEPDisabled;
	}

	tmp = cpu_to_le32(auth);
	rndis_set_oid(usbdev, OID_802_11_AUTHENTICATION_MODE, &tmp, sizeof(tmp));
	tmp = cpu_to_le32(enc);
	rndis_set_oid(usbdev, OID_802_11_ENCRYPTION_STATUS, &tmp, sizeof(tmp));

out:
	kfree(wepkey);
	return ret;
}


static int rndis_iw_set_encode_ext(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct iw_encode_ext * ext = (void*)extra;
	struct usbnet * usbdev = dev->priv;
	int ret = -EINVAL, keyindex;
	NDIS_802_11_REMOVE_KEY rem_key;
	NDIS_802_11_KEY add_key;

	devdbg(usbdev, "SIOCSIWENCODEEXT: ext. flags:%08x, alg:%08x: keylen:%d",
				ext->ext_flags, ext->alg, ext->key_len);
	devdbg(usbdev, "                : enc. flags:%08x, length:%d",
				wrqu->encoding.flags, wrqu->encoding.length);

	keyindex = (wrqu->encoding.flags & IW_ENCODE_INDEX) - 1;

	if ( ext->ext_flags & IW_ENCODE_EXT_SET_TX_KEY)
		keyindex |= (1 << 31);  // TX key

	if ( ! (ext->ext_flags & IW_ENCODE_EXT_GROUP_KEY) )
		keyindex |= (1 << 30);  // pairwise key

	if (wrqu->encoding.flags & IW_ENCODE_DISABLED) {
		memset(&rem_key, 0, sizeof(rem_key));
		rem_key.Length = cpu_to_le32(sizeof(rem_key));
		rem_key.KeyIndex = cpu_to_le32(keyindex);
		if (keyindex & (1 << 30)) {
			memcpy(rem_key.BSSID, ext->addr.sa_data, ETH_ALEN);
		} else {
			memset(rem_key.BSSID, 0xff, ETH_ALEN);
		}
		ret = rndis_set_oid(usbdev, OID_802_11_REMOVE_KEY, &rem_key,
							sizeof(rem_key));
	} else {
		if (ext->key_len > sizeof(add_key.KeyMaterial))
			return -EINVAL;

		memset(&add_key, 0, sizeof(add_key));
		add_key.Length = cpu_to_le32(sizeof(add_key));

		if (ext->ext_flags & IW_ENCODE_EXT_RX_SEQ_VALID) {
			memcpy(&add_key.KeyRSC, ext->rx_seq,
					IW_ENCODE_SEQ_MAX_SIZE);
			keyindex |= (1 << 29);
		}

		if (ext->ext_flags & IW_ENCODE_EXT_TX_SEQ_VALID) {
			memcpy(&add_key.KeyRSC, ext->tx_seq,
					IW_ENCODE_SEQ_MAX_SIZE);
			keyindex |= (1 << 29);
		}

		add_key.KeyIndex = cpu_to_le32(keyindex);
		add_key.KeyLength = cpu_to_le32(ext->key_len);
		if (keyindex & (1 << 30)) {
			memcpy(add_key.BSSID, ext->addr.sa_data, ETH_ALEN);
		} else {
			memset(add_key.BSSID, 0xff, ETH_ALEN);
		}
		if (ext->alg == IW_ENCODE_ALG_TKIP) {
			memcpy(add_key.KeyMaterial, ext->key, 16);
			memcpy(add_key.KeyMaterial + 16, ext->key + 24, 8);
			memcpy(add_key.KeyMaterial + 24, ext->key + 16, 8);
		} else {
			memcpy(add_key.KeyMaterial, ext->key, ext->key_len);
		}
		// kludge alert - breaking a race between ioctl and network
		// packet.  This delay is to prevent encryption being enabled
		// while the last authentication handshake packet is still
		// underway in the stack somewhere. If encryption is enabled
		// too early the AP will not see this last handshake packet and
		// association fails.
		//
		// also see:
		// http://lists.shmoo.com/pipermail/hostap/2006-September/014157.html
		//
		msleep(500);

		ret = rndis_set_oid(usbdev, OID_802_11_ADD_KEY, &add_key,
						    sizeof(add_key));
	}
	return ret;
}


static int rndis_iw_set_scan(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct iw_param * param = &wrqu->param;
	struct usbnet * usbdev = dev->priv;
	union iwreq_data evt;
	int ret = -EINVAL;
	__le32 tmp;

	devdbg(usbdev, "SIOCSIWSCAN");

	if (param->flags == 0) {
		tmp = ccpu2(1);
		ret = rndis_set_oid(usbdev, OID_802_11_BSSID_LIST_SCAN, &tmp,
								sizeof(tmp));
		evt.data.flags = 0;
		evt.data.length = 0;
		wireless_send_event(dev, SIOCGIWSCAN, &evt, NULL);
	}
	return ret;
}


static int rndis_iw_get_scan(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	struct iw_event iwe;
	char *cev = extra;
	char *end_buf = extra + IW_SCAN_MAX_DATA;
	void *buf = NULL;
	NDIS_802_11_BSSID_LIST_EX *bssid_list;
	NDIS_WLAN_BSSID_EX *bssid;
	int ret = -EINVAL, len, count;

	devdbg(usbdev, "SIOCGIWSCAN");

	buf = kmalloc(CONTROL_BUFFER_SIZE, GFP_KERNEL);
	if (!buf) {
		ret = -ENOMEM;
		goto out;
	}

	len = CONTROL_BUFFER_SIZE;
	ret = rndis_query_oid(usbdev, OID_802_11_BSSID_LIST, buf, &len);

	if (ret != 0)
		goto out;

#ifdef DEBUG
	//rndis_dump_buffer(usbdev, "BSSID scan data", buf, len);
#endif

	bssid_list = buf;
	bssid = bssid_list->Bssid;
	count = le32_to_cpu(bssid_list->NumberOfItems);
	devdbg(usbdev, "SIOCGIWSCAN: %d BSSIDs found", count);

	while (count &&
		((void*)bssid + le32_to_cpu(bssid->Length) - 1) < (buf + len)) {
		devdbg(usbdev, "BSSID %02x:%02x:%02x:%02x:%02x:%02x",
				bssid->MacAddress[0], bssid->MacAddress[1],
				bssid->MacAddress[2], bssid->MacAddress[3],
				bssid->MacAddress[4], bssid->MacAddress[5]);
		iwe.cmd = SIOCGIWAP;
		iwe.u.ap_addr.sa_family = ARPHRD_ETHER;
		memcpy(iwe.u.ap_addr.sa_data, bssid->MacAddress, ETH_ALEN);
		cev = iwe_stream_add_event(cev, end_buf, &iwe, IW_EV_ADDR_LEN);

		devdbg(usbdev, "SSID(%d) %s",
			le32_to_cpu(bssid->Ssid.SsidLength),
			bssid->Ssid.Ssid);
		iwe.cmd = SIOCGIWESSID;
		iwe.u.essid.length = le32_to_cpu(bssid->Ssid.SsidLength);
		iwe.u.essid.flags = 1;
		cev = iwe_stream_add_point(cev, end_buf, &iwe,
							bssid->Ssid.Ssid);

		devdbg(usbdev, "MODE %d",
				le32_to_cpu(bssid->InfrastructureMode));
		iwe.cmd = SIOCGIWMODE;
		switch (le32_to_cpu(bssid->InfrastructureMode)) {
		case Ndis802_11IBSS:
			iwe.u.mode = IW_MODE_ADHOC;
			break;
		case Ndis802_11Infrastructure:
			iwe.u.mode = IW_MODE_INFRA;
			break;
		case Ndis802_11AutoUnknown:
			iwe.u.mode = IW_MODE_AUTO;
			break;
		}
		cev = iwe_stream_add_event(cev, end_buf, &iwe, IW_EV_UINT_LEN);

		devdbg(usbdev, "FREQ %d kHz",
			le32_to_cpu(bssid->Configuration.DSConfig));
		iwe.cmd = SIOCGIWFREQ;
		iwe.u.freq.m = le32_to_cpu(bssid->Configuration.DSConfig);
		iwe.u.freq.e = 3;
		iwe.u.freq.flags = 0;
		cev = iwe_stream_add_event(cev, end_buf, &iwe, IW_EV_FREQ_LEN);

		devdbg(usbdev, "QUAL %d",
						le32_to_cpu(bssid->Rssi));
		iwe.cmd = IWEVQUAL;
		iwe.u.qual.qual = le32_to_cpu(bssid->Rssi);
		iwe.u.qual.level = le32_to_cpu(bssid->Rssi);
		iwe.u.qual.noise = 0;
		iwe.u.qual.updated = IW_QUAL_QUAL_UPDATED
				| IW_QUAL_LEVEL_UPDATED
				| IW_QUAL_DBM
				| IW_QUAL_NOISE_INVALID;
		cev = iwe_stream_add_event(cev, end_buf, &iwe, IW_EV_QUAL_LEN);

		devdbg(usbdev, "ENCODE %d",
						le32_to_cpu(bssid->Privacy));
		iwe.cmd = SIOCGIWENCODE;
		iwe.u.data.length = 0;
		if (le32_to_cpu(bssid->Privacy))
			iwe.u.data.flags = IW_ENCODE_ENABLED;
		else
			iwe.u.data.flags = IW_ENCODE_DISABLED;
		cev = iwe_stream_add_point(cev, end_buf, &iwe, NULL);

		devdbg(usbdev, "IEs");
		iwe.cmd = IWEVGENIE;
		iwe.u.data.length = le32_to_cpu(bssid->IELength) -
						sizeof(NDIS_802_11_FIXED_IEs);
		iwe.u.data.flags = 1;
		cev = iwe_stream_add_point(cev, end_buf, &iwe,
			(char*)bssid->IEs + sizeof(NDIS_802_11_FIXED_IEs));

		bssid = (void*)bssid + le32_to_cpu(bssid->Length);
		count--;
	}

out:
	wrqu->data.length = cev - extra;
	wrqu->data.flags = 0;
	if (buf)
		kfree(buf);
	return ret;
}


static int rndis_iw_set_mlme(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct iw_mlme * mlme = (struct iw_mlme*)extra;
	struct usbnet * usbdev = dev->priv;
	int ret = -ENOTSUPP;
	__le32 tmp;

#ifdef DEBUG
	u8 *bssid = (u8*)mlme->addr.sa_data;
	devdbg(usbdev, "SIOCSIWMLME: %04x, %04x, %02x:%02x:%02x:%02x:%02x:%02x",
			mlme->cmd, mlme->reason_code, bssid[0], bssid[1],
			bssid[2], bssid[3], bssid[4], bssid[5]);
#endif

	switch (mlme->cmd) {
	case IW_MLME_DEAUTH:
		// fallthrough ok ??
	case IW_MLME_DISASSOC:
		tmp = ccpu2(1);
		ret = rndis_set_oid(usbdev, OID_802_11_DISASSOCIATE, &tmp,
								sizeof(tmp));
		break;
	}

	return ret;
}


static int rndis_iw_set_genie(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	struct rndis_private * priv = usbdev->rndis_wext_priv;
	int ret = 0;

#ifdef DEBUG
	int j;
	u8 * gie = extra;
	for (j=0; j < wrqu->data.length; j+=8) {
		devdbg(usbdev,
			"SIOCSIWGENIE %04x - "
			"%02x %02x %02x %02x %02x %02x %02x %02x", j,
			gie[j + 0], gie[j + 1], gie[j + 2], gie[j + 3],
		   	gie[j + 4], gie[j + 5], gie[j + 6], gie[j + 7]);
	}
#endif
	// clear existing IEs
	if (priv->wpa_ie_len) {
		kfree(priv->wpa_ie);
		priv->wpa_ie_len = 0;
	}

	// set new IEs
	priv->wpa_ie = kmalloc(wrqu->data.length, GFP_KERNEL);
	if (priv->wpa_ie) {
		priv->wpa_ie_len = wrqu->data.length;
		memcpy(priv->wpa_ie, extra, priv->wpa_ie_len);
	} else
		ret = -ENOMEM;
	return ret;
}


static int rndis_iw_set_rts(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	__le32 tmp;
	devdbg(usbdev, "SIOCSIWRTS");

	tmp = cpu_to_le32(wrqu->rts.value);
	return rndis_set_oid(usbdev, OID_802_11_RTS_THRESHOLD, &tmp,
								sizeof(tmp));
}


static int rndis_iw_get_rts(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	__le32 tmp;
	int len, ret;
	devdbg(usbdev, "SIOCGIWRTS");

	len = sizeof(tmp);
	ret = rndis_query_oid(usbdev, OID_802_11_RTS_THRESHOLD, &tmp, &len);
	if (ret == 0) {
		wrqu->rts.value = le32_to_cpu(tmp);
		wrqu->rts.flags = 1;
		wrqu->rts.disabled = 0;
	}
	return ret;
}

static int rndis_iw_set_frag(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	__le32 tmp;
	devdbg(usbdev, "SIOCSIWFRAG");

	tmp = cpu_to_le32(wrqu->frag.value);
	return rndis_set_oid(usbdev, OID_802_11_FRAGMENTATION_THRESHOLD, &tmp,
								sizeof(tmp));
}


static int rndis_iw_get_frag(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	__le32 tmp;
	int len,ret;
	devdbg(usbdev, "SIOCGIWFRAG");

	len = sizeof(tmp);
	ret = rndis_query_oid(usbdev, OID_802_11_FRAGMENTATION_THRESHOLD, &tmp,
									&len);
	if (ret == 0) {
		wrqu->frag.value = le32_to_cpu(tmp);
		wrqu->frag.flags = 1;
		wrqu->frag.disabled = 0;
	}
	return ret;
}


static int rndis_iw_set_nick(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	struct rndis_private * priv = usbdev->rndis_wext_priv;

	priv->nick_len = wrqu->data.length;
	if (priv->nick_len > 32)
		priv->nick_len = 32;

	memcpy(priv->nick, extra, priv->nick_len);
	return 0;
}


static int rndis_iw_get_nick(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	struct rndis_private * priv = usbdev->rndis_wext_priv;

	wrqu->data.flags = 1;
	wrqu->data.length = priv->nick_len;
	memcpy(extra, priv->nick, priv->nick_len);
	return 0;
}


static int rndis_iw_set_freq(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	// TBD
	// fake success to make iwconfig continue
	return 0;
}


static int rndis_iw_get_freq(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	NDIS_802_11_CONFIGURATION config;
	int len, ret;

	len = sizeof(config);
	ret = rndis_query_oid(usbdev, OID_802_11_CONFIGURATION, &config, &len);
	if (ret == 0) {
		wrqu->freq.m = le32_to_cpu(config.DSConfig); // freq in kHz
		wrqu->freq.e = 3;
		wrqu->freq.flags = 0;
	}
	return ret;
}


static int rndis_iw_get_txpower(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	__le32 tmp;
	int ret, len;

	len = sizeof(tmp);
	ret = rndis_query_oid(usbdev, OID_802_11_TX_POWER_LEVEL, &tmp, &len);
	if (ret == 0) {
		wrqu->txpower.flags = IW_TXPOW_MWATT;
		wrqu->txpower.value = le32_to_cpu(tmp);
		wrqu->txpower.disabled = 0;
	}
	return ret;
}


static int rndis_iw_get_rate(struct net_device *dev,
    struct iw_request_info *info, union iwreq_data * wrqu, char *extra)
{
	struct usbnet * usbdev = dev->priv;
	__le32 tmp;
	int ret, len;

	len = sizeof(tmp);
	ret = rndis_query_oid(usbdev, OID_GEN_LINK_SPEED, &tmp, &len);
	if (ret == 0) {
		wrqu->bitrate.value = le32_to_cpu(tmp) * 100;
		wrqu->bitrate.disabled = 0;
		wrqu->bitrate.flags = 1;
	}
	return ret;
}


static struct iw_statistics * rndis_get_wireless_stats(
						struct net_device * dev)
{
	struct usbnet * usbdev = dev->priv;
	struct rndis_private * priv = usbdev->rndis_wext_priv;
	__le32 rssi;
	int len,ret;

	memset(&priv->iwstats, 0, sizeof(priv->iwstats));

	/* cannot do this here since rndis_command sleeps */

	len = sizeof(rssi);
//	ret = rndis_query_oid(usbdev, OID_802_11_RSSI, &rssi, &len);

	ret = 0;
	rssi = cpu_to_le32(-30);
	
	if (ret == 0) {
		priv->iwstats.qual.qual  = le32_to_cpu(rssi);
		priv->iwstats.qual.level = le32_to_cpu(rssi);
		priv->iwstats.qual.updated = IW_QUAL_QUAL_UPDATED
				| IW_QUAL_LEVEL_UPDATED
				| IW_QUAL_DBM
				| IW_QUAL_NOISE_INVALID;
	}

	return &priv->iwstats;
}


#define IW_IOCTL(x) [(x) - SIOCSIWCOMMIT]
static const iw_handler rndis_iw_handler[] =
{
	IW_IOCTL(SIOCSIWCOMMIT)    = rndis_iw_commit,
	IW_IOCTL(SIOCGIWNAME)      = rndis_iw_get_name,
	IW_IOCTL(SIOCSIWFREQ)      = rndis_iw_set_freq,
	IW_IOCTL(SIOCGIWFREQ)      = rndis_iw_get_freq,
	IW_IOCTL(SIOCSIWMODE)      = rndis_iw_set_mode,
	IW_IOCTL(SIOCGIWMODE)      = rndis_iw_get_mode,
	IW_IOCTL(SIOCGIWRANGE)     = rndis_iw_get_range,
	IW_IOCTL(SIOCSIWAP)        = rndis_iw_set_bssid,
	IW_IOCTL(SIOCGIWAP)        = rndis_iw_get_bssid,
	IW_IOCTL(SIOCSIWSCAN)      = rndis_iw_set_scan,
	IW_IOCTL(SIOCGIWSCAN)      = rndis_iw_get_scan,
	IW_IOCTL(SIOCSIWESSID)     = rndis_iw_set_essid,
	IW_IOCTL(SIOCGIWESSID)     = rndis_iw_get_essid,
	IW_IOCTL(SIOCSIWNICKN)     = rndis_iw_set_nick,
	IW_IOCTL(SIOCGIWNICKN)     = rndis_iw_get_nick,
	IW_IOCTL(SIOCGIWRATE)      = rndis_iw_get_rate,
	IW_IOCTL(SIOCSIWRTS)       = rndis_iw_set_rts,
	IW_IOCTL(SIOCGIWRTS)       = rndis_iw_get_rts,
	IW_IOCTL(SIOCSIWFRAG)      = rndis_iw_set_frag,
	IW_IOCTL(SIOCGIWFRAG)      = rndis_iw_get_frag,
	IW_IOCTL(SIOCGIWTXPOW)     = rndis_iw_get_txpower,
	IW_IOCTL(SIOCSIWENCODE)    = rndis_iw_set_encode,
	IW_IOCTL(SIOCSIWENCODEEXT) = rndis_iw_set_encode_ext,
	IW_IOCTL(SIOCSIWMLME)      = rndis_iw_set_mlme,
	IW_IOCTL(SIOCSIWAUTH)      = rndis_iw_set_auth,
	IW_IOCTL(SIOCSIWGENIE)     = rndis_iw_set_genie,
};

static const iw_handler rndis_private_handler[] = {
};

static const struct iw_priv_args rndis_private_args[] = {
};


static const struct iw_handler_def rndis_iw_handlers = {
	.num_standard = ARRAY_SIZE(rndis_iw_handler),
	.num_private  = ARRAY_SIZE(rndis_private_handler),
	.num_private_args = ARRAY_SIZE(rndis_private_args),
	.standard = (iw_handler *)rndis_iw_handler,
	.private  = (iw_handler *)rndis_private_handler,
	.private_args = (struct iw_priv_args *)rndis_private_args,
	.get_wireless_stats = rndis_get_wireless_stats,
};


void rndis_wext_link_change(struct usbnet *dev, int state)
{
	struct rndis_private *priv = dev->rndis_wext_priv;
	union iwreq_data evt;
	u8 bssid[ETH_ALEN];

	if (state) {
		int ret, len = ETH_ALEN;
		ret = rndis_query_oid(dev, OID_802_11_BSSID, bssid, &len);
		if (ret == 0)
			memcpy(priv->bssid, bssid, ETH_ALEN);
		memcpy(evt.ap_addr.sa_data, priv->bssid, ETH_ALEN);
	} else {
		memset(priv->bssid, 0, ETH_ALEN);
		memset(evt.ap_addr.sa_data, 0, ETH_ALEN);
	}
	evt.ap_addr.sa_family = ARPHRD_ETHER;
	wireless_send_event(dev->net, SIOCGIWAP, &evt, NULL);
}


static int rndis_wext_get_caps(struct usbnet *dev)
{
	struct {
		__le32	num_items;
		__le32	items[8];
	} networks_supported;
	int len, retval, i, n;
	struct rndis_private * priv = dev->rndis_wext_priv;

	// determine supported modes
	len = sizeof(networks_supported);
	retval = rndis_query_oid(dev, OID_802_11_NETWORK_TYPES_SUPPORTED,
						&networks_supported, &len);
	if (retval >= 0) {
		n = le32_to_cpu(networks_supported.num_items);
		if (n > 8)
		    n = 8;
		for (i = 0; i < n; i++) {
			switch (le32_to_cpu(networks_supported.items[i])) {
			case Ndis802_11FH:
			case Ndis802_11DS:
				priv->caps |= CAP_MODE_80211B;
				break;
			case Ndis802_11OFDM5:
				priv->caps |= CAP_MODE_80211A;
				break;
			case Ndis802_11OFDM24:
				priv->caps |= CAP_MODE_80211G;
				break;
			}
		}
		if (priv->caps & CAP_MODE_80211A)
			strcat(priv->name, "a");
		if (priv->caps & CAP_MODE_80211B)
			strcat(priv->name, "b");
		if (priv->caps & CAP_MODE_80211G)
			strcat(priv->name, "g");
	}

	return retval;
}


int rndis_wext_probe(struct usbnet *dev, struct usb_interface *intf)
{
	u32 oid;
	u32 *oids;
	int i, count, len;
	int retval = 0;

	/* get supported OIDs */
	oids = kmalloc(CONTROL_BUFFER_SIZE, GFP_KERNEL);
	if (!oids) {
		retval = -ENOMEM;
		goto out;
	}

	len = CONTROL_BUFFER_SIZE;
	rndis_query_oid(dev, OID_GEN_SUPPORTED_LIST, oids, &len);

	/* run through list and check for 802.11 OIDS (0x0d in MSB) */
	i = 0;
	count = len / sizeof(u32);
	while (i < count) {
		oid = le32_to_cpu(oids[i]);
		if ((oid >> 24) == 0x0d) {
			retval = 1;
			break;
		}
		i++;
	}

out:
	return retval;
}


int rndis_wext_bound(struct usbnet *dev)
{
	return (dev->net->wireless_handlers != NULL);
}


int rndis_wext_bind(struct usbnet *dev, struct usb_interface *intf)
{
	struct net_device *net = dev->net;
	struct rndis_private * priv;
	int retval = 0;
	__le32 tmp;

	/* allocate rndis private data */
	dev->rndis_wext_priv = kmalloc(sizeof(struct rndis_private), GFP_KERNEL);
	if (!dev->rndis_wext_priv)
		return -ENOMEM;

	priv = dev->rndis_wext_priv;
	memset(priv, 0, sizeof(*priv));
	memset(priv->name, 0, sizeof(priv->name));
	strcpy(priv->name, "IEEE802.11");
	net->wireless_handlers = &rndis_iw_handlers;

	rndis_wext_get_caps(dev);

	tmp = cpu_to_le32(Ndis802_11WEPDisabled);
	rndis_set_oid(dev, OID_802_11_ENCRYPTION_STATUS, &tmp,sizeof(tmp));
	tmp = cpu_to_le32(Ndis802_11AuthModeOpen);
	rndis_set_oid(dev, OID_802_11_AUTHENTICATION_MODE, &tmp, sizeof(tmp));
	tmp = cpu_to_le32(Ndis802_11PrivFilter8021xWEP);
	rndis_set_oid(dev, OID_802_11_PRIVACY_FILTER, &tmp, sizeof(tmp));

	return retval;
}


void rndis_wext_unbind(struct usbnet *dev, struct usb_interface *intf)
{
	struct rndis_private * priv = dev->rndis_wext_priv;

	if (priv && priv->wpa_ie_len)
	    kfree(priv->wpa_ie);
	kfree(dev->rndis_wext_priv);
}

